# the semantic version scheme here is read as follows:
# The major version: Only incremented on major breaking changes
# The minor version: Incremented upon each minor release.
# The patch version: Corresponds to Trino's grammar version.
__version__ = "0.0.368"



"""
Every relation in Treeno is a :class:`Relation`. `Relations
<https://en.wikipedia.org/wiki/Relation_(database)>`_ are SQL entities that contain rows of data,
with every row being the same data type, and each element in the row are assigned its corresponding column name.
You can think of relations as "things I can FROM in a SELECT query".

For example, the following constructs are all relations:

.. code-block:: sql

    trino> VALUES 1,2,3;
     _col0
    -------
         1
         2
         3

    trino> SELECT 1,2,3;
     _col0 | _col1 | _col2
    -------+-------+-------
         1 |     2 |     3

    trino> SELECT * FROM UNNEST(ARRAY[1,2,3]);
     _col0
    -------
         1
         2
         3

    trino> SELECT * FROM some_table;
     _col0
    -------
         1
         2
         3

Not all :class:`Relation` s are directly queryable though! For those that are directly queryable, i.e. SELECT statements,
TABLE statements, etc, they inherit from :class:`Query` which directly inherits from :class:`Relation`.

Since Treeno does not try to communicate with the metastore(s) powering :class:`Table` s, we need user input to resolve
schemas. More information can be found in :class:`Relation`.
"""

import copy
from abc import ABC, abstractmethod
from enum import Enum
from typing import Dict, List, Optional, Set, Type

import attr

from treeno.base import PrintMode, PrintOptions, SetQuantifier, Sql
from treeno.datatypes import types as type_consts
from treeno.datatypes.builder import row, unknown
from treeno.datatypes.conversions import common_supertype
from treeno.datatypes.types import DataType
from treeno.expression import Value, value_attr
from treeno.groupby import GroupBy
from treeno.orderby import OrderTerm
from treeno.printer import StatementPrinter, join_stmts, pad
from treeno.util import chain_identifiers, parenthesize, quote_identifier
from treeno.window import Window


class Relation(Sql, ABC):
    """Represents a SQL relation.

    A :class:`Relation` can be one of the following:

    1. (:class:`Table`) A table reference.
    2. (:class:`Query`) A subquery. :class:`Query` is an abstract class, and its subclasses contain
        :class:`SelectQuery`, :class:`ValuesTable`, :class:`SetQuery` and its subclasses, etc.
    3. (:class:`Unnest`) An unnested expression of arrays.
    4. (:class:`Lateral`) Shares the current relation's namespace with any :class:`Query`.
    5. (:class:`TableSample`) A subset sample of any relation.
    6. (:class:`Join`) A join composing of any relation.
    7. (:class:`AliasedRelation`) An alias (both as relation name and its column names) of any relation.
    """

    def identifier(self) -> Optional[str]:
        """Shorthand identifier for the relation

        .. todo:: This identifier function is used to check for whether dereferences in fields correspond to a relation,
            but there are multiple ways we can dereference. catalog.schema.table.field is a valid identifier, not only
            table.field, which identifier would return "table" for and match on.

        Returns:
            A string if the relation has a well-defined identifier, otherwise None
        """
        return None

    @abstractmethod
    def resolve(self, existing_schema: "Schema") -> "Schema":
        """Resolves the current relation's schema with supplemental schema information.

        This function is used to resolve missing types in queries. The user can pass in existing_schemas to
        hint at types of dynamically determined fields i.e. a table in Trino, but we also use resolve() underneath
        the hood in order to resolve some types that can only be determined through a cross-query tree traversal. For
        more information refer to :class:`Schema`.

        Args:
            existing_schema: Schema with a partial list of fields to resolve the current field's schema
        Returns:
            A new schema for the current relation
        """
        ...


@attr.s
class SchemaField:
    """Represents a single field in a :class:`Schema`

    A schema must have a data type, a source, and optionally have a name. For example, a SUM(x) term with no alias
    should have no name. It still shows up in the output, but there's no way to refer to it.

    Attributes:
        name: An optional name of the schema field. If the name is None, then the field cannot be selected later on,
            and Trino will autogenerate column names for these e.g. _col0
        source: The source of which the field came from. Important to disambiguate :class:`SchemaField` with same
            names across different sources
        data_type: The data type of the field
    """

    name: Optional[str] = attr.ib()
    source: Relation = attr.ib()
    data_type: DataType = attr.ib()


@attr.s
class Schema:
    """Represents the output schema of a given :class:`Relation`

    :py:class:`Schema` s are used to impute missing types from trees with partial type information. This class serves two
    main functions. The first is to resolve cross-query references i.e. WITH queries which can be referenced from the
    FROM clause in a :class:`SelectQuery`. The second is to allow the user to pass in schemas to resolve relations
    like :class:`Table` s which can't resolve its own schema without access to external metadata (i.e. from a metastore).

    Attributes:
        fields: A list of :class:`SchemaField` s that exist in the current namespace
        Relation_ids: A set of relations used to denote the existence of a relation included in the schema.
            This set can contain relations that contains no fields - e.g. for a table with an undefined schema,
            the fields would be empty but the relation would still be in the set
    """

    fields: List[SchemaField] = attr.ib()
    relation_ids: Set[str] = attr.ib()

    @classmethod
    def empty_schema(cls) -> "Schema":
        return cls(fields=[], relation_ids=set())

    def merge(self, another_schema: "Schema") -> "Schema":
        """Merge two schemas together to create a new schema

        >>> from treeno.datatypes.builder import bigint
        >>> schema_a = Schema([SchemaField("x", Table("a"), bigint())], relation_ids={"a"})
        >>> schema_b = Schema([SchemaField("y", Table("b"), bigint())], relation_ids={"b"})
        >>> resulting_schema = schema_a.merge(schema_b)
        >>> assert resulting_schema.fields == schema_a.fields + schema_b.fields
        >>> assert resulting_schema.relation_ids == {"a", "b"}

        Args:
            another_schema: The other schema to merge with. The fields from it are appended to the end of the resulting
                :class:`Schema` fields
        Returns:
            A new :class:`Schema` object with both schemas' fields and relations in the same namespace.
        """
        # .. todo:: Should this be deep copies? I don't see a point in trying to deepcopy relations since they're expensive.
        fields = copy.copy(self.fields)
        relation_ids = copy.copy(self.relation_ids)
        # .. todo:: This is also pretty slow, O(N^2)
        for f in another_schema.fields:
            if f not in fields:
                fields.append(f)
        relation_ids |= another_schema.relation_ids
        return Schema(fields, relation_ids)


@value_attr
class Query(Relation, Value, ABC):
    """Represents a query with filtered outputs.

    Note that Queries are also :class:`treeno.expression.Values`, in that they are just row data typed expressions.
    """

    # .. todo:: Technically, an offset can either be an integer or a question mark(as a parameter).
    offset: Optional[int] = attr.ib(default=None, kw_only=True)
    limit: Optional[int] = attr.ib(default=None, kw_only=True)
    orderby: Optional[List[OrderTerm]] = attr.ib(default=None, kw_only=True)
    with_: List["AliasedRelation"] = attr.ib(factory=dict, kw_only=True)

    def _with_query_string_builder(self, opts: PrintOptions) -> Dict[str, str]:
        """Creates a dictionary mapping of the WITH clause to the queries.

        This function is used solely for :func:`treeno.Sql.sql`, where we use :class:`treeno.printer.StatementPrinter`
        to format the "river" for readability.
        """
        if not self.with_:
            return {}
        # NOTE: We hold AliasedRelations in self.with_, but the sql output for it should be in
        # cte namedQuery form - it's not the same as a typical AliasedRelation.sql().
        return {
            "WITH": join_stmts(
                [query._named_query_sql(opts) for query in self.with_], opts
            )
        }

    def _resolve_with(self) -> Schema:
        """Resolves the queries in WITH statement with context that carries over.

        It's important to note that CTE's can actually refer to previously defined CTE's using FROM:

        .. code-block:: sql

            trino> WITH a (foo) AS (SELECT 1), b (bar) AS (SELECT a.foo + 1 FROM a) SELECT bar FROM b;
             bar
            -----
               2
            (1 row)

        In the above example, ``b`` is able to refer to ``a`` because it's defined before in the WITH sequence.
        """
        merging_schema = Schema.empty_schema()
        for w in self.with_:
            new_schema = w.resolve(merging_schema)
            merging_schema = merging_schema.merge(new_schema)
        return merging_schema

    def _constraint_string_builder(self, opts: PrintOptions) -> Dict[str, str]:
        """Creates a dictionary mapping of a couple output-filtering constraints to the query.

        This function is used solely for :func:`treeno.Sql.sql`, where we use :class:`treeno.printer.StatementPrinter`
        to format the "river" for readability.
        """
        str_builder = {}
        if self.orderby:
            # Typically the BY in ORDER BY clause goes across the indentation river.
            # .. todo:: The "BY" will offset the max length slightly - we might see join_stmts max out at 80 characters
            #       but we also have a few characters added due to the BY.
            str_builder["ORDER"] = "BY " + join_stmts(
                [order.sql(opts) for order in self.orderby], opts
            )
        if self.offset:
            str_builder["OFFSET"] = str(self.offset)
        if self.limit:
            str_builder["LIMIT"] = str(self.limit)
        return str_builder


@value_attr
class SetQuery(Query, ABC):
    """Represents a set operation on two subqueries.

    For all set operations, the input query schemas must be the coercible with each other (i.e. integer and bigint).
    Otherwise, Trino will complain:

    .. code-block:: sql

        trino> SELECT 1 UNION SELECT 'a';
        ... column 1 in UNION query has incompatible types: integer, varchar(1)

    Attributes:
        left_query: An arbitrary :class:`Query`. The set query's output will be identical in schema to this query.
        right_query: An arbitrary :class:`Query` to filter ``left_query`` on.
        set_quantifier: Quantifier for the output rows. If DISTINCT, then the output rows will only contain unique rows.
            If ALL, then the output rows will contain all rows of the set operation.
    """

    left_query: Query = attr.ib()
    right_query: Query = attr.ib()
    # NOTE: Set quantifier for set queries is DISTINCT by default, instead of ALL.
    set_quantifier: SetQuantifier = attr.ib(default=SetQuantifier.DISTINCT)

    def __attrs_post_init__(self) -> None:
        self.data_type = self._compute_data_type()

    def _compute_data_type(self) -> DataType:
        left_type, right_type = (
            self.left_query.data_type,
            self.right_query.data_type,
        )
        if left_type == unknown() or right_type == unknown():
            return unknown()

        # The subqueries not being row-like is possible for a table/subquery that only has 1 column
        left_row_like = left_type.type_name == type_consts.ROW
        right_row_like = right_type.type_name == type_consts.ROW
        assert (
            left_row_like == right_row_like
        ), f"Either both types are ROWs or neither are. Found {left_type} and {right_type}"

        if not left_row_like:
            return common_supertype(left_type, right_type)

        # .. todo:: Redundant after previous assertion
        assert (
            left_row_like and right_row_like
        ), "Input data types for query set operations must be ROW"
        dtypes = []
        for t1, t2 in zip(
            left_type.parameters["dtypes"], right_type.parameters["dtypes"]
        ):
            dtypes.append(common_supertype(t1, t2))
        return row(dtypes=dtypes)

    def _to_string(self, set_type: str, opts: PrintOptions) -> str:
        spacing = "\n" if opts.mode == PrintMode.PRETTY else " "
        query_string = relation_string(self.left_query, opts, newline=False)
        query_string += spacing
        query_string += set_type
        # NOTE: We currently have set_quantifier explicitly spelled out for set operations because the default ALL
        # for expressions doesn't apply here(the default here is DISTINCT).
        query_string += f" {self.set_quantifier.name}"
        query_string += spacing
        query_string += relation_string(self.right_query, opts, newline=False)
        return query_string

    def resolve(self, existing_schema: Schema) -> Schema:
        left_schema = self.left_query.resolve(
            maybe_prune_schema(self.left_query, existing_schema)
        )
        self.right_query.resolve(
            maybe_prune_schema(self.right_query, existing_schema)
        )
        # The column names and such are always taken from the left schema.
        schema_fields = []
        for schema_field in left_schema.fields:
            # Basically take the schema that belonged to left schema and re-assign it to this set query
            schema_fields.append(
                SchemaField(schema_field.name, self, schema_field.data_type)
            )
        return Schema(schema_fields, relation_ids=left_schema.relation_ids)


@value_attr
class IntersectQuery(SetQuery):
    """Represents an intersection of two queries.

    Note that intersect queries cannot use set quantifier ALL, since Trino implements INTERSECT by hashing the rows
    and subsequently deduplicates input query rows.

    >>> from treeno.expression import NULL
    >>> q = SelectQuery([NULL])
    >>> str(IntersectQuery(q, q, set_quantifier=SetQuantifier.DISTINCT))
    '(SELECT NULL) INTERSECT DISTINCT (SELECT NULL)'
    >>> IntersectQuery(q, q, set_quantifier=SetQuantifier.ALL)
    Traceback (most recent call last):
        ...
    AssertionError: INTERSECT does not support ALL
    """

    def __attrs_post_init__(self) -> None:
        assert (
            self.set_quantifier == SetQuantifier.DISTINCT
        ), f"INTERSECT does not support {self.set_quantifier.name}"

    def sql(self, opts: PrintOptions) -> str:
        return self._to_string("INTERSECT", opts)


@value_attr
class ExceptQuery(SetQuery):
    """Represents a set subtract of left query with right query.

    Note that except queries cannot use set quantifier ALL, since Trino implements EXCEPT by hashing the rows
    and subsequently deduplicates input query rows.

    >>> from treeno.expression import NULL
    >>> q = SelectQuery([NULL])
    >>> str(ExceptQuery(q, q, set_quantifier=SetQuantifier.DISTINCT))
    '(SELECT NULL) EXCEPT DISTINCT (SELECT NULL)'
    >>> ExceptQuery(q, q, set_quantifier=SetQuantifier.ALL)
    Traceback (most recent call last):
        ...
    AssertionError: EXCEPT does not support ALL
    """

    def __attrs_post_init__(self) -> None:
        assert (
            self.set_quantifier == SetQuantifier.DISTINCT
        ), f"EXCEPT does not support {self.set_quantifier.name}"

    def sql(self, opts: PrintOptions) -> str:
        return self._to_string("EXCEPT", opts)


@value_attr
class UnionQuery(SetQuery):
    """Represents a union of two queries' rows.

    >>> from treeno.expression import NULL
    >>> q = SelectQuery([NULL])
    >>> str(UnionQuery(q, q, set_quantifier=SetQuantifier.DISTINCT))
    '(SELECT NULL) UNION DISTINCT (SELECT NULL)'
    >>> str(UnionQuery(q, q, set_quantifier=SetQuantifier.ALL))
    '(SELECT NULL) UNION ALL (SELECT NULL)'
    """

    def sql(self, opts: PrintOptions) -> str:
        return self._to_string("UNION", opts)


@value_attr
class SelectQuery(Query):
    """Represents a high level SELECT query.

    >>> from treeno.expression import wrap_literal, AliasedValue, Field
    >>> table = Table("a")
    >>> query = SelectQuery(
    ...     select=[AliasedValue(wrap_literal(2), "foo"), Field("a") / 5],
    ...     from_=table,
    ...     where=Field("a") > 5
    ... )
    >>> # We can get the SQL string of the query via __str__
    >>> str(query)
    'SELECT 2 "foo","a" / 5 FROM "a" WHERE "a" > 5'
    >>> # The schema of the fields can be retrieved from resolve
    >>> query_fields = query.resolve(Schema.empty_schema()).fields
    >>> # We aliased the literal value as "foo", but we didn't give a name for the second field
    >>> [schema_field.name for schema_field in query_fields]
    ['foo', None]
    >>> # The type for field a is unknown without a schema supplied to the Table object.
    >>> [str(schema_field.data_type) for schema_field in query_fields]
    ['INTEGER', 'UNKNOWN']
    >>> # Alternatively, we can also get the whole row type of the select query via:
    >>> str(query.data_type)
    'ROW(INTEGER,UNKNOWN)'

    Attributes:
        select: A list of :class:`Value` s as outputs to the query.
        from_: An optional relation to select from. If from is not specified, select must contain values that do not
            reference any external relations, i.e. ``SELECT 1``.
        where: An optional boolean clause to filter rows by. If where is not specified, all rows are used.
        groupby: An optional :class:`GroupBy` which allows the user to run partial aggregate functions over different
            groupings. For more information refer to :mod:`treeno.groupby`.
        having: An optional boolean clause to filter groups by. If having is not specified, all groups are used.
        select_quantifier: Whether to return all rows or only the distinct ones. Defaults to ALL.
        window: An optional mapping of window names to :class:`Window` s. For more information refer to
            :mod:`treeno.window`.
    """

    select: List[Value] = attr.ib()
    from_: Optional[Relation] = attr.ib(default=None)
    where: Optional[Value] = attr.ib(default=None)
    groupby: Optional[GroupBy] = attr.ib(default=None)
    having: Optional[Value] = attr.ib(default=None)
    select_quantifier: SetQuantifier = attr.ib(factory=SetQuantifier.default)
    window: Optional[Dict[str, Window]] = attr.ib(default=None, kw_only=True)

    def __attrs_post_init__(self) -> None:
        assert len(
            self.select
        ), "Must select at least one value for SELECT clause"
        self.data_type = self._compute_data_type()

    def _compute_data_type(self) -> DataType:
        # Special case: If the select clause only has one output element, it's not a row.
        if len(self.select) == 1:
            return self.select[0].data_type
        return row(dtypes=[val.data_type for val in self.select])

    def resolve(self, existing_schema: Schema) -> Schema:
        from treeno.datatypes.resolve import resolve_fields

        # .. todo:: Currently resolve only handles a single layer. We should make this a common function across all
        #       relations so we can recursively call resolve before performing resolve on this query.
        # Also, it's probably worth allowing table inputs.
        # Pass in existing relations from CTE into from_ as long as from_ is not in its own namespace i.e. a subquery.
        if self.from_ is not None:
            existing_schema = existing_schema.merge(self._resolve_with())
            schema = self.from_.resolve(
                maybe_prune_schema(self.from_, existing_schema)
            )
            self.select = [resolve_fields(val, schema) for val in self.select]
            self.data_type = self._compute_data_type()

        schema_fields = []
        for value in self.select:
            # All schema fields have the source to the current object
            schema_fields.append(
                SchemaField(
                    name=value.identifier(),
                    source=self,
                    data_type=value.data_type,
                )
            )
        return Schema(schema_fields, relation_ids=set())

    def sql(self, opts: PrintOptions) -> str:
        builder = StatementPrinter()
        builder.update(self._with_query_string_builder(opts))
        select_value = join_stmts([val.sql(opts) for val in self.select], opts)
        # All is the default, so we don't need to mention it
        if self.select_quantifier != SetQuantifier.default():
            select_value = self.select_quantifier.name + " " + select_value
        builder.add_entry("SELECT", select_value)

        if self.from_:
            relation_str = self.from_.sql(opts)
            # Queries need to be parenthesized to be considered relations
            if isinstance(self.from_, Query):
                # Add padding of 1 character to realign the statement)
                relation_str = pad(parenthesize(relation_str), 1)
            builder.add_entry("FROM", relation_str)
        if self.where:
            builder.add_entry("WHERE", self.where.sql(opts))
        if self.groupby:
            builder.add_entry("GROUP", "BY " + self.groupby.sql(opts))
        if self.having:
            builder.add_entry("HAVING", self.having.sql(opts))
        if self.window:
            window_string = join_stmts(
                [
                    f"{window_name} AS {parenthesize(window.sql(opts))}"
                    for window_name, window in self.window.items()
                ],
                opts,
            )
            builder.add_entry("WINDOW", window_string)
        builder.update(self._constraint_string_builder(opts))
        return builder.to_string(opts)


@attr.s
class Table(Relation):
    """A table reference uniquely identified by a qualified name

    Tables can be standalone queries by themselves (in the form TABLE {table name}), and thus
    can be subject to orderby, offset, and limit constraints. Refer to TableQuery for more information.

    >>> from treeno.datatypes.builder import bigint
    >>> table = Table("a")
    >>> assert table._column_schema == Schema([], relation_ids={"a"})
    >>> # Supply a schema to the table via resolve
    >>> new_schema = table.resolve(Schema([SchemaField("x", table, bigint())], relation_ids={"a"}))
    >>> schema_fields = new_schema.fields
    >>> assert table._column_schema == new_schema
    >>> [schema_field.name for schema_field in schema_fields]
    ['x']
    >>> [str(schema_field.data_type) for schema_field in schema_fields]
    ['BIGINT']

    Attributes:
        name: The name of the table.
        schema: The schema the table belongs to. Can be unspecified to denote the current session's schema.
        catalog: The catalog the schema belongs to. Can be unspecified to denote the current session's catalog.
    """

    name: str = attr.ib()
    schema: Optional[str] = attr.ib(default=None)
    catalog: Optional[str] = attr.ib(default=None)

    # The table object will remember its own column_schema here.
    _column_schema: Schema = attr.ib(factory=Schema.empty_schema)

    def __attrs_post_init__(self) -> None:
        if self.catalog:
            assert (
                self.schema
            ), "If a catalog is specified, a schema must be specified as well"
        if self.name not in self._column_schema.relation_ids:
            self._column_schema.relation_ids.add(self.name)

    def sql(self, opts: PrintOptions) -> str:
        return chain_identifiers(self.catalog, self.schema, self.name)

    def resolve(self, existing_schema: Schema) -> Schema:
        # If this schema isn't defined
        if self.name not in existing_schema.relation_ids:
            return self._column_schema
        # Otherwise, the schema IS defined e.g. by a previous CTE
        schema_fields = [
            f
            for f in existing_schema.fields
            if f.source.identifier() == self.identifier()
        ]
        self._column_schema = Schema(
            schema_fields, relation_ids={self.identifier()}
        )
        return self._column_schema

    def identifier(self) -> Optional[str]:
        return self.name


@value_attr
class TableQuery(Query):
    """A light wrapper around a table for querying.

    >>> str(TableQuery(Table("x")))
    'TABLE "x"'

    The above statement is equivalent to SELECT * FROM x.

    Attributes:
        table: The underlying table to select from.
    """

    table: Table = attr.ib()

    def __attrs_post_init__(self) -> None:
        self.data_type = self._compute_data_type()

    def _compute_data_type(self) -> DataType:
        if self.table._column_schema == Schema.empty_schema():
            return row(dtypes=[f.data_type for f in self.table._column_schema])
        return unknown()

    def sql(self, opts: PrintOptions) -> str:
        builder = StatementPrinter()
        builder.update(self._with_query_string_builder(opts))
        builder.add_entry("TABLE", self.table.sql(opts))
        builder.update(self._constraint_string_builder(opts))
        return builder.to_string(opts)

    def resolve(self, existing_schema: Schema) -> Schema:
        # We not only need to take the existing schema, but we also need to shadow some fields with the
        # local WITH clause
        with_schema = self._resolve_with()
        table_schema = self.table.resolve(with_schema)
        self.data_type = self._compute_data_type()
        return table_schema


@value_attr
class ValuesQuery(Query):
    """A literal table constructed by literal ROWs

    For a given comma separated list of expressions, the VALUES keyword turns the expression into a subquery that can
    be used later on. VALUES is often used for unit testing fixtures for mock tables.

    Values tables can be standalone queries by themselves, and thus can be subject to
    orderby, offset, and limit constraints. Refer to ValuesQuery for more information.

    >>> from decimal import Decimal
    >>> from treeno.expression import wrap_literal, RowConstructor
    >>> # Types are inferred from all input values coerced together
    >>> query = ValuesQuery([wrap_literal(1), wrap_literal(Decimal("2.2"))])
    >>> str(query)
    'VALUES 1,2.2'
    >>> str(query.data_type)
    'DECIMAL(11,1)'
    >>> query = ValuesQuery([
    ...     RowConstructor([
    ...         wrap_literal(1),
    ...         wrap_literal('a')
    ...     ]),
    ...     RowConstructor([
    ...         wrap_literal(2),
    ...         wrap_literal('b')
    ...     ])
    ... ])
    >>> str(query)
    "VALUES (1,'a'),(2,'b')"
    >>> str(query.data_type)
    'ROW(INTEGER,VARCHAR(1))'

    Attributes:
        exprs: Values for each row of the query.
    """

    exprs: List[Value] = attr.ib()

    def __attrs_post_init__(self) -> None:
        for val in self.exprs:
            if self.data_type == unknown():
                self.data_type = val.data_type
            else:
                self.data_type = common_supertype(self.data_type, val.data_type)

    def sql(self, opts: PrintOptions) -> str:
        builder = StatementPrinter()
        builder.update(self._with_query_string_builder(opts))
        builder.add_entry(
            "VALUES", join_stmts([expr.sql(opts) for expr in self.exprs], opts)
        )
        builder.update(self._constraint_string_builder(opts))
        return builder.to_string(opts)

    def resolve(self, existing_schema: Schema) -> Schema:
        # ..todo:: Does ValuesQuery dereference anything by itself?
        return Schema(
            [SchemaField(None, self, val.data_type) for val in self.exprs]
        )


@attr.s
class AliasedRelation(Relation):
    """Represents an alias corresponding to a relation

    Aliased relations can change/add identifier to another underlying :class:`Relation`, and can also rname the columns.
    Here's an example using resolve:

    >>> from treeno.expression import wrap_literal
    >>> aliased_relation = AliasedRelation(SelectQuery([wrap_literal(1), wrap_literal('2')]), "query", ["x", "y"])
    >>> aliased_schema = aliased_relation.resolve(Schema.empty_schema())
    >>> [schema_field.name for schema_field in aliased_schema.fields]
    ['x', 'y']
    >>> aliased_schema.relation_ids
    {'query'}

    .. todo:: Currently we don't check to see whether the field length is equal to our column_aliases parameter. Should
        we check this or let it fail later?

    Attributes:
        relation: A relation to alias over.
        alias: Alias for the relation.
        column_aliases: An optional list of names for each column of the underlying relation.
    """

    relation: Relation = attr.ib()
    alias: str = attr.ib()
    column_aliases: Optional[List[str]] = attr.ib(default=None)

    def sql(self, opts: PrintOptions) -> str:
        # .. todo:: Keep the "AS"?
        alias_str = f"{relation_string(self.relation, opts)} {quote_identifier(self.alias)}"
        if self.column_aliases:
            alias_str += f" ({join_stmts(self.column_aliases, opts)})"
        return alias_str

    def _named_query_sql(self, opts: PrintOptions) -> str:
        alias_str = quote_identifier(self.alias)
        if self.column_aliases:
            alias_str += f" {join_stmts(self.column_aliases, opts)}"
        alias_str += f" AS {relation_string(self.relation, opts)}"
        return alias_str

    def resolve(self, existing_schema: Schema) -> Schema:
        relation_schema = self.relation.resolve(
            maybe_prune_schema(self.relation, existing_schema)
        )
        new_schema_fields = []
        for idx, schema_field in enumerate(relation_schema.fields):
            field_name = (
                self.column_aliases[idx]
                if self.column_aliases
                else schema_field.name
            )
            new_schema_fields.append(
                SchemaField(field_name, self, schema_field.data_type)
            )
        return Schema(new_schema_fields, relation_ids={self.identifier()})

    def identifier(self) -> Optional[str]:
        return self.alias


class JoinType(Enum):
    INNER = "INNER"
    LEFT = "LEFT"
    RIGHT = "RIGHT"
    OUTER = "FULL OUTER"
    CROSS = "CROSS"


class JoinCriteria(Sql, ABC):
    """Join criterias are complex expressions that describe exactly how a JOIN is done."""

    @abstractmethod
    def build_sql(self, opts: PrintOptions) -> Dict[str, str]:
        """Creates a dictionary mapping of statements to strings
        """
        ...

    def sql(self, opts: PrintOptions) -> str:
        return StatementPrinter(stmt_mapping=self.build_sql(opts)).to_string(
            opts
        )


@attr.s
class JoinUsingCriteria(JoinCriteria):
    """Using allows us to join on equality criterion on multiple rows.

    There's one subtle difference between USING and ON, which is the output number of columns:

    .. code-block:: sql

        SELECT * FROM (SELECT 1 AS "foo") JOIN (SELECT 1 AS "foo") USING("foo");
         foo
        -----
           1
        (1 row)

    Selects the column once. If we use ON:

    .. code-block:: sql

        SELECT * FROM (SELECT 1 AS "foo") "a" JOIN (SELECT 1 AS "foo") "b" ON "a"."foo" = "b"."foo";
         foo | foo
        -----+-----
           1 |   1
        (1 row)

    If the output columns have the same name, then upon referencing the columns Trino will fail.

    Attributes:
        column_names: A list of column names to check equality clauses for.
    """

    # NOTE: We can't use Field from treeno.expression since they refer to a single relation. These column names refer to
    # both the left and right relations of the join.
    column_names: List[str] = attr.ib()

    def build_sql(self, opts: PrintOptions) -> Dict[str, str]:
        return {"USING": parenthesize(join_stmts(self.column_names, opts))}


@attr.s
class JoinOnCriteria(JoinCriteria):
    """Perform a join between two relations using boolean expressions.

    An example ON usage:

    .. code-block:: sql

        SELECT * FROM (SELECT 1 AS "foo") "a" JOIN (SELECT 1 AS "foo") "b" ON "a"."foo" = "b"."foo";
         foo | foo
        -----+-----
           1 |   1
        (1 row)

    Attributes:
        constraint: A boolean expression constraining the join to where the expression evaluates to True using
            two subqueries' rows.
    """

    constraint: Value = attr.ib()

    def build_sql(self, opts: PrintOptions) -> Dict[str, str]:
        # For complex boolean expressions i.e. conjunctions and disjunctions we have a new line, so we have to
        # indent it here for readability
        return {"ON": pad(self.constraint.sql(opts), opts.spaces)}


@attr.s
class JoinConfig(Sql):
    """Details the method of join used in a :class:`Join`.

    Attributes:
        join_type: The type of JOIN (i.e. CROSS JOIN, INNER JOIN, etc)
        natural: Whether the join is natural (i.e. joins on equality on column names that are the same between two queries)
        criteria: What boolean expression criteria the join requires. Refer to :class:`JoinCriteria` for more info.
    """

    join_type: JoinType = attr.ib()
    # Natural joins are dangerous and should be avoided, but since
    # it's valid Trino grammar we'll allow it for now.
    natural: bool = attr.ib(default=False)
    criteria: Optional[JoinCriteria] = attr.ib(default=None)

    def __attrs_post_init__(self):
        if self.join_type == JoinType.CROSS:
            assert not self.criteria, "Cross joins cannot specify join criteria"

        if self.criteria:
            assert (
                not self.natural
            ), "If criteria is specified, the join cannot be natural"

    def sql(self, opts: PrintOptions) -> str:
        raise NotImplementedError(
            "JoinConfig.sql should not be used. Refer to Join"
        )


@attr.s
class Join(Relation):
    """Represents a join between two relations

    Attributes:
        left_relation: An arbitrary relation to be joined on
        right_relation: An arbitrary relation to be joined on
        config: Details on what type of join we're performing. Refer to :class:`JoinConfig` for more info.
    """

    left_relation: Relation = attr.ib()
    right_relation: Relation = attr.ib()
    config: JoinConfig = attr.ib()

    def sql(self, opts: PrintOptions) -> str:
        builder = StatementPrinter(river=False)
        # No value to the key, which is just the relation itself
        builder.add_entry(relation_string(self.left_relation, opts), "")
        join_type = ""
        if self.config.natural:
            join_type += "NATURAL "
        join_type += f"{self.config.join_type.value} JOIN"
        builder.add_entry(join_type, relation_string(self.right_relation, opts))
        if self.config.criteria is not None:
            builder.update(self.config.criteria.build_sql(opts))
        return builder.to_string(opts)

    def resolve(self, existing_schema: Schema) -> Schema:
        pruned_left_arg = maybe_prune_schema(
            self.left_relation, existing_schema
        )
        left_schema = self.left_relation.resolve(pruned_left_arg)
        existing_schema = existing_schema.merge(left_schema)
        # Pass in the namespace of the first relation to the second. This is important for sql statemnts like
        # SELECT x FROM a CROSS JOIN UNNEST(a.foo), where UNNEST requires a to be selected first
        pruned_right_arg = maybe_prune_schema(
            self.right_relation, existing_schema
        )
        right_schema = self.right_relation.resolve(pruned_right_arg)
        return left_schema.merge(right_schema)


@attr.s
class Unnest(Relation):
    """Represents an unnested set of arrays representing a table

    An example of unnest:

    .. code-block:: sql

        trino> SELECT * FROM UNNEST(ARRAY[1,2]);
         _col0
        -------
             1
             2

    Attributes:
        arrays: The array(s) to use in an unnest. The i-th array in the list corresponds to the i-th output column.
        with_ordinality: Whether there should be an ordinality column appended at the end of the UNNEST.
    """

    arrays: List[Value] = attr.ib()
    with_ordinality: bool = attr.ib(default=False, kw_only=True)

    def __attrs_post_init__(self) -> None:
        self.data_type = self._compute_data_type()

    def _compute_data_type(self) -> DataType:
        dtypes = []
        for val in self.arrays:
            if val.data_type.type_name != type_consts.ARRAY:
                dtypes.append(unknown())
            else:
                dtypes.append(val.data_type.parameters["dtype"])
        return row(dtypes=dtypes)

    def sql(self, opts: PrintOptions) -> str:
        str_builder = [
            f"UNNEST({join_stmts([arr.sql(opts) for arr in self.arrays], opts)})"
        ]
        if self.with_ordinality:
            str_builder += " WITH ORDINALITY"
        return " ".join(str_builder)

    def resolve(self, existing_schema: Schema) -> Schema:
        from treeno.datatypes.resolve import resolve_fields

        # There's no base
        self.arrays = [
            resolve_fields(arr, existing_schema) for arr in self.arrays
        ]
        self.data_type = self._compute_data_type()
        return Schema(
            fields=[
                SchemaField(name=None, source=self, data_type=dtype)
                for dtype in self.data_type.parameters["dtypes"]
            ],
            relation_ids=set(),
        )


@attr.s
class Lateral(Relation):
    """Represents a correlated subquery.

    Attributes:
        subquery: The subquery to correlate with the current scope.
    """

    subquery: Query = attr.ib()

    def sql(self, opts: PrintOptions) -> str:
        return f"LATERAL({self.subquery.sql(opts)})"

    def resolve(self, existing_schema: Schema) -> Schema:
        # NOTE: We explicitly don't maybe_prune_schema here, since lateral needs to explicitly pass it down
        return self.subquery.resolve(existing_schema)


class SampleType(Enum):
    BERNOULLI = "BERNOULLI"
    SYSTEM = "SYSTEM"


@attr.s
class TableSample(Relation):
    """Represents a sampled relation.

    Note that TABLESAMPLE can work with all relations, including JOINs. However, for JOINs we need to take extra care
    to parenthesize the expression. This doesn't work:

    .. code-block:: sql

        trino> WITH f (a) AS (SELECT 1), g (b) AS (SELECT 1) SELECT f.a, g.b FROM f JOIN g ON f.a = g.b TABLESAMPLE BERNOULLI(99);
        ... mismatched input 'TABLESAMPLE' ...

    But this does:

    .. code-block:: sql

        trino> WITH f (a) AS (SELECT 1), g (b) AS (SELECT 1) SELECT f.a, g.b FROM (f JOIN g ON f.a = g.b) TABLESAMPLE BERNOULLI(99);
         a | b
        ---+---
         1 | 1

    Because JOINs are higher on the parsing hierarchy than TABLESAMPLEs, potentially due to an ambiguous parse
    otherwise where TABLESAMPLE can be applied to the right relation in a JOIN.
    """

    relation: Relation = attr.ib()
    sample_type: SampleType = attr.ib()
    percentage: Value = attr.ib()

    def sql(self, opts: PrintOptions) -> str:
        # Queries need to be parenthesized to be considered relations
        relation_sql = relation_string(
            self.relation,
            opts,
            newline=False,
            special_parenthesize_relations=[Join],
        )
        return f"{relation_sql} TABLESAMPLE {self.sample_type.value}({self.percentage.sql(opts)})"

    def resolve(self, existing_schema: Schema) -> Schema:
        # NOTE: We don't maybe_prune_schema here, because table sample should not affect the namespace scope
        return self.relation.resolve(existing_schema)


def relation_string(
    relation: Relation,
    opts: PrintOptions,
    newline: bool = True,
    special_parenthesize_relations: Optional[List[Type[Relation]]] = None,
) -> str:
    relation_str = relation.sql(opts)
    parenthesize_classes = [Query]
    if special_parenthesize_relations:
        parenthesize_classes.extend(special_parenthesize_relations)
    if isinstance(relation, tuple(parenthesize_classes)):
        if opts.mode == PrintMode.PRETTY and newline:
            relation_str = "\n" + relation_str
        return parenthesize(relation_str)
    else:
        return relation_str


def maybe_prune_schema(relation: Relation, existing_schema: Schema) -> Schema:
    """Prune the schema depending on whether the relation will enter a new namespace

    Some relations don't want extra schema information to be passed through because it's in its own parenthesized
    expression which is not aware of the outer namespace.

    Args:
        relation: The relation to enter into with the given schema
        existing_schema: The schema to pass into the relation
    Returns:
        Either an empty :class:`Schema` or the original :class:`Schema`.
    """
    if isinstance(relation, Query):
        return Schema.empty_schema()
    return existing_schema



from enum import Enum
from io import StringIO
from typing import Any

import typer
from nltk.tree import Tree

from treeno.base import DefaultableEnum, PrintMode, PrintOptions, Sql
from treeno.builder.convert import (
    expression_from_sql,
    query_from_sql,
    type_from_sql,
)
from treeno.datatypes.builder import unknown
from treeno.datatypes.types import DataType
from treeno.grammar.parse import AST
from treeno.grammar.parse import tree as print_tree
from treeno.relation import Schema
from treeno.util import children, is_dictlike, is_listlike

app = typer.Typer()


class SqlConstruct(str, Enum):
    QUERY = "query"
    EXPRESSION = "expression"
    TYPE = "type"


def get_sql_object(construct_type: SqlConstruct, sql: str) -> Sql:
    sql_object: Sql
    if construct_type == SqlConstruct.QUERY:
        sql_object = query_from_sql(sql)
        # Resolve a query since it is much more complicated
        sql_object.resolve(Schema.empty_schema())
    elif construct_type == SqlConstruct.EXPRESSION:
        sql_object = expression_from_sql(sql)
    elif construct_type == SqlConstruct.TYPE:
        sql_object = type_from_sql(sql)
    else:
        raise TypeError(f"Unexpected sql construct type {construct_type.value}")
    return sql_object


@app.command()
def format(construct_type: SqlConstruct, sql: str) -> None:
    sql_obj = get_sql_object(construct_type, sql)
    typer.echo(sql_obj.sql(PrintOptions(PrintMode.PRETTY)))


@app.command()
def antlr_tree(construct_type: SqlConstruct, sql: str) -> None:
    ast = AST(sql)
    if construct_type == SqlConstruct.QUERY:
        node = ast.query()
    elif construct_type == SqlConstruct.EXPRESSION:
        node = ast.expression()
    elif construct_type == SqlConstruct.TYPE:
        node = ast.type()
    else:
        raise TypeError(f"Unexpected sql construct type {construct_type.value}")
    typer.echo(print_tree(ast, node))


def _should_skip(key: str, node: Any) -> bool:
    # Ignore hidden members
    if key.startswith("_"):
        return True
    if node is None:
        return True
    if isinstance(node, DefaultableEnum) and node == type(node).default():
        return True
    if (is_listlike(node) or is_dictlike(node)) and not len(node):
        return True
    return False


def treeify(node: Any) -> Any:
    if is_listlike(node):
        return Tree(
            f"<{type(node).__name__}>", [treeify(item) for item in node]
        )
    if is_dictlike(node):
        return Tree(
            f"<{type(node).__name__}>",
            [Tree(k, treeify(v)) for k, v in node.items()],
        )
    if not isinstance(node, Sql):
        return str(node)
    # Data types are terminal and we don't want to print out the parameters when there is none
    if isinstance(node, DataType):
        return str(node)
    child_nodes = [
        Tree(k, [treeify(v)])
        for k, v in children(node).items()
        if not _should_skip(k, v)
    ]
    return Tree(node.__class__.__name__, child_nodes)


@app.command()
def tree(construct_type: SqlConstruct, sql: str, draw: bool = False) -> None:
    sql_obj = get_sql_object(construct_type, sql)
    tree = treeify(sql_obj)
    sio = StringIO()
    if not draw:
        tree.pretty_print(stream=sio)
        typer.echo(sio.getvalue())
    else:
        tree.draw()


@app.command()
def schema(sql: str) -> None:
    sql_obj = get_sql_object(SqlConstruct.QUERY, sql)
    schema = sql_obj.resolve(Schema.empty_schema())
    if not schema.fields:
        typer.echo(
            typer.style("Unknown schema", fg=typer.colors.RED, bold=True)
        )
        return
    lines = []
    for idx, field in enumerate(schema.fields):
        field_string = (
            typer.style("<none>", fg=typer.colors.RED, bold=True)
            if field is None
            else field.name
        )
        dtype_string = (
            typer.style("<unknown>", fg=typer.colors.RED, bold=True)
            if field.data_type == unknown()
            else field.data_type
        )
        lines.append(f"{field_string} - {dtype_string}")
    typer.echo("\n".join(lines))



from enum import auto

import attr

from treeno.base import DefaultableEnum, GenericEnum, PrintOptions, Sql
from treeno.expression import Value


class OrderType(DefaultableEnum):
    ASC = auto()
    DESC = auto()

    @classmethod
    def default(cls: GenericEnum) -> GenericEnum:
        return cls.ASC


class NullOrder(DefaultableEnum):
    FIRST = auto()
    LAST = auto()

    @classmethod
    def default(cls: GenericEnum) -> GenericEnum:
        return cls.LAST


@attr.s
class OrderTerm(Sql):
    value: Value = attr.ib()
    order_type: OrderType = attr.ib(factory=OrderType.default)
    null_order: NullOrder = attr.ib(factory=NullOrder.default)

    def sql(self, opts: PrintOptions):
        order_string = self.value.sql(opts)
        if self.order_type != OrderType.default():
            order_string += f" {self.order_type.name}"
        if self.null_order != NullOrder.default():
            order_string += f" NULLS {self.null_order.name}"
        return order_string



from abc import ABC
from typing import List

import attr

from treeno.base import PrintOptions, SetQuantifier, Sql
from treeno.expression import Value
from treeno.printer import join_stmts


class Group(Sql, ABC):
    """Base class to describe all grouping entities.
    """


@attr.s
class GroupBy(Sql):
    """GroupBys are used to group rows by their membership in grouping sets to get partial aggregates.
    """

    groups: List[Group] = attr.ib()
    groupby_quantifier: SetQuantifier = attr.ib(factory=SetQuantifier.default)

    def sql(self, opts: PrintOptions) -> str:
        groupby_string = join_stmts(
            [group.sql(opts) for group in self.groups], opts
        )
        if self.groupby_quantifier != SetQuantifier.default():
            groupby_string = f"{self.groupby_quantifier.name} {groupby_string}"
        return groupby_string


@attr.s
class GroupingSet(Group):
    """Simple group involving input expr(s). Note that this doesn't accept aliased values, since group bys directly
    reference the input column.

    Note that the grouping set can be empty - it would correspond to interpreting the entire table as one group.
    """

    values: List[Value] = attr.ib()

    def sql(self, opts: PrintOptions) -> str:
        if len(self.values) == 1:
            return self.values[0].sql(opts)
        return f"({join_stmts([val.sql(opts) for val in self.values], opts)})"


@attr.s
class GroupingSetList(Group):
    """Describes multiple groupings.
    NOTE: We explicitly disallow the feature of grouping by an index here. This is not a TODO.
    TODO: Give this a better name
    """

    groups: List[GroupingSet] = attr.ib()

    def __attrs_post_init__(self) -> None:
        assert (
            len(self.groups) > 0
        ), "GroupingSetLists must have at least one grouping set"

    def sql(self, opts: PrintOptions) -> str:
        return f"GROUPING SETS ({join_stmts([group.sql(opts) for group in self.groups], opts)})"


@attr.s
class Cube(Group):
    """Creates a grouping set of the powerset of all fields

    For example, CUBE(a,b) yields the grouping sets (), (a), (b), (a,b)
    """

    values: List[Value] = attr.ib()

    def __attrs_post_init__(self) -> None:
        assert (
            len(self.values) > 0
        ), "Cube must have at least one value to group by"

    def sql(self, opts: PrintOptions) -> str:
        return (
            f"CUBE ({join_stmts([val.sql(opts) for val in self.values], opts)})"
        )


@attr.s
class Rollup(Group):
    """Creates a grouping set of cumulatively aggregated columns in order of input.

    For example, ROLLUP(a,b) yields the grouping sets (), (a), (a,b)
    """

    values: List[Value] = attr.ib()

    def __attrs_post_init__(self) -> None:
        assert (
            len(self.values) > 0
        ), "GroupingSets must have at least one value to group by"

    def sql(self, opts: PrintOptions) -> str:
        return f"ROLLUP ({join_stmts([val.sql(opts) for val in self.values], opts)})"



import itertools
from typing import Any, Dict, Iterable, Iterator, List, Optional, TypeVar

import attr

from treeno.base import Sql

T = TypeVar("T")


def quote_identifier(identifier: str) -> str:
    return f'"{identifier}"'


def parenthesize(val: Any) -> str:
    return f"({val})"


def chain_identifiers(*identifiers: Optional[str], join_string=".") -> str:
    """Chains a list of identifiers together by periods or whatever join_string is
    For example, chain_identifier(None, None, "a", "b") will give us "a"."b"
    """
    not_null_identifiers: List[str] = [
        identifier for identifier in identifiers if identifier is not None
    ]
    return join_string.join(
        [quote_identifier(identifier) for identifier in not_null_identifiers]
    )


def quote_literal(literal: str) -> str:
    return f"'{literal}'"


def nth(
    iterable: Iterable[T], n: int, default: Optional[T] = None
) -> Optional[T]:
    "Returns the nth item or a default value"
    return next(itertools.islice(iterable, n, None), default)


def is_listlike(var: Any) -> bool:
    return isinstance(var, (tuple, list, set, frozenset))


def is_dictlike(var: Any) -> bool:
    return isinstance(var, dict)


def construct_container(var: Iterable[T], it: Iterator[T]) -> Iterable[T]:
    return type(var)(it)


def children(sql: Sql, recursive: bool = False) -> Dict[str, Any]:
    fields_dict = attr.fields_dict(type(sql))
    return {field_name: getattr(sql, field_name) for field_name in fields_dict}



from abc import ABC
from enum import Enum
from typing import List, Optional

import attr

from treeno.base import DefaultableEnum, GenericEnum, PrintOptions, Sql
from treeno.expression import Value, wrap_literal
from treeno.orderby import OrderTerm
from treeno.printer import StatementPrinter, join_stmts


class NullTreatment(DefaultableEnum):
    """Used in conjunction with window functions in primary expressions"""

    IGNORE = "IGNORE"
    RESPECT = "RESPECT"

    @classmethod
    def default(cls: GenericEnum) -> GenericEnum:
        return NullTreatment.RESPECT


class FrameType(DefaultableEnum):
    RANGE = "RANGE"
    ROWS = "ROWS"
    GROUPS = "GROUPS"

    @classmethod
    def default(cls: GenericEnum) -> GenericEnum:
        return FrameType.RANGE


class BoundType(Enum):
    PRECEDING = "PRECEDING"
    FOLLOWING = "FOLLOWING"


class FrameBound(Sql, ABC):
    """Represents a bound in the range of a window frame specification"""


@attr.s
class BoundedFrameBound(FrameBound):
    bound_type: BoundType = attr.ib()
    offset: Value = attr.ib(converter=wrap_literal)

    def sql(self, opts: PrintOptions) -> str:
        return f"{self.offset.sql(opts)} {self.bound_type.value}"


@attr.s
class UnboundedFrameBound(FrameBound):
    bound_type: BoundType = attr.ib()

    def sql(self, opts: PrintOptions) -> str:
        return f"UNBOUNDED {self.bound_type.value}"


@attr.s
class CurrentFrameBound(FrameBound):
    def sql(self, opts: PrintOptions) -> str:
        return "CURRENT ROW"


def default_start_bound() -> UnboundedFrameBound:
    return UnboundedFrameBound(bound_type=BoundType.PRECEDING)


def default_end_bound() -> CurrentFrameBound:
    return CurrentFrameBound()


@attr.s
class Window(Sql):
    # TODO: Evaluate whether it's worth it to embed a parent window class here.
    # The grammar makes it hard to embed a linked list data structure of window references during parse.
    parent_window: Optional[str] = attr.ib(default=None)
    orderby: Optional[List[OrderTerm]] = attr.ib(default=None)
    partitions: Optional[List[Value]] = attr.ib(default=None)
    frame_type: FrameType = attr.ib(factory=FrameType.default)
    # TODO: For now we represent missing bounds as default bounds
    start_bound: FrameBound = attr.ib(factory=default_start_bound)
    end_bound: FrameBound = attr.ib(factory=default_end_bound)

    def sql(self, opts: PrintOptions) -> str:
        builder = StatementPrinter()
        if self.parent_window:
            # Empty string for no value on the right side of the river
            builder.add_entry(self.parent_window, "")
        if self.partitions:
            partition_value = join_stmts(
                [partition.sql(opts) for partition in self.partitions], opts
            )
            builder.add_entry("PARTITION", "BY " + partition_value)
        if self.orderby:
            order_value = join_stmts(
                [order_term.sql(opts) for order_term in self.orderby], opts
            )
            builder.add_entry("ORDER", "BY " + order_value)
        builder.add_entry(
            self.frame_type.value,
            " ".join(
                [
                    "BETWEEN",
                    self.start_bound.sql(opts),
                    "AND",
                    self.end_bound.sql(opts),
                ]
            ),
        )
        return builder.to_string(opts)



"""
Treeno is built around the concept of creating python nodes that represent SQL constructs. These constructs must be
interoperable with SQL strings, which means:

1. The trees created by a hierarchy of nodes can produce valid SQL.
2. We can construct a Treeno tree from valid SQL.

In order to do this, this module introduces some preliminary classes for printing/formatting SQL and the base
:class:`Sql` superclass responsible for representing all SQL nodes.

.. todo:: We currently copy this description into api.rst, because we don't want to include all the
"""

from abc import ABC, ABCMeta, abstractmethod
from enum import Enum, EnumMeta, auto
from typing import Any, TypeVar

import attr

GenericEnum = TypeVar("GenericEnum", bound=Enum)


class ABCEnumMeta(EnumMeta, ABCMeta):
    """This is required because Enum classes have a different metaclass, so we must merge both enum and abc metaclasses
    into a single class"""

    ...


class DefaultableEnum(Enum, metaclass=ABCEnumMeta):
    @classmethod
    @abstractmethod
    def default(cls: GenericEnum) -> GenericEnum:
        raise NotImplementedError(
            f"All {cls.__name__} must implement default()"
        )


class PrintMode(DefaultableEnum):
    DEFAULT = auto()
    PRETTY = auto()

    @classmethod
    def default(cls: GenericEnum) -> GenericEnum:
        return cls.DEFAULT


@attr.s
class PrintOptions:
    mode: PrintMode = attr.ib(factory=PrintMode.default)
    spaces: int = attr.ib(default=4)


@attr.s
class Sql(ABC):
    """A base class for all SQL nodes in Treeno."""

    @abstractmethod
    def sql(self, opts: PrintOptions) -> str:
        """Converts the Treeno object into SQL string.

        All objects that represent some concept in SQL inherit from :class:`~treeno.base.Sql` and must
        implement this function to return its SQL representation which may be more compact
        or more readable depending on the :class:`~treeno.base.PrintOptions`. By default, all Treeno
        objects has a ``__str__`` representation which calls this function with default
        print options.

        When writing a new class that inherits from :class:`~treeno.base.Sql`, remember to pass along
        ``opts`` to its childrens' ``sql()`` call to make sure the print options are
        recursively applied.

        >>> from treeno.expression import wrap_literal, AliasedValue, Field
        >>> from treeno.relation import SelectQuery, Table
        >>> table = Table("a")
        >>> query = SelectQuery(select=[AliasedValue(wrap_literal(2), "foo"), Field("a") / 5], from_=table, where=Field("a") > 5)
        >>> print(query.sql(PrintOptions(mode=PrintMode.DEFAULT)))
        SELECT 2 "foo","a" / 5 FROM "a" WHERE "a" > 5
        >>> print(query.sql(PrintOptions(mode=PrintMode.PRETTY, spaces=2)))
        SELECT 2 "foo","a" / 5
          FROM "a"
         WHERE "a" > 5

        Args:
            opts: The print options to control the SQL output format.
        Returns:
            A SQL string for the given object.
        """
        raise NotImplementedError(
            f"All {self.__class__.__name__} must implement sql"
        )

    def __str__(self) -> str:
        # Default print options
        return self.sql(PrintOptions())

    def equals(self, other: Any) -> bool:
        """Checks to see whether two :class:`Sql` nodes have identical content.

        Note:
            Why do we not use ``__eq__`` here? Because for the sake of syntactical sugar, we need that operator for
            :class:`~treeno.expression.Value` to evaluate whether two objects are equal in SQL space (thus generating
            a Sql node using :class:`~treeno.expression.Equal` rather than returning to us a boolean value).

        Args:
            other: Another potentially SQL-like object.
        Returns:
            True if both objects are equal, False otherwise.
        """
        if not isinstance(other, Sql) or type(self) is not type(other):
            return False
        self_dict = attr.asdict(self)
        other_dict = attr.asdict(other)
        return self_dict == other_dict

    def assert_equals(self, other: Any) -> None:
        """Assert whether two :class:`Sql` nodes are the same.

        For more information, refer to :func:`Sql.equals`.
        """
        assert self.equals(other)


class SetQuantifier(DefaultableEnum):
    """Whether to select all rows or only distinct values
    """

    DISTINCT = auto()
    ALL = auto()

    @classmethod
    def default(cls: GenericEnum) -> GenericEnum:
        return cls.ALL



"""
Treeno supports arbitrary expressions including arithmetic binary, arithmetic unary, function, boolean expressions and
more. Underneath the hood, the values to be expressed in SQL are actually nodes of a tree and the tree is traversed to
generate SQL for execution. The base class which supports a suite of syntactic sugar such as ``__add__``, ``__eq__`` and
is :class:`Value`.

Here's an example using the CLI tool:

.. code-block:: bash

    ❯ treeno tree expression "1+2*3"
                                         Add
         _________________________________|__________________________
        |                  |                                       right
        |                  |                                         |
        |                  |                                      Multiply
        |                  |               __________________________|__________________
        |                 left            |                 left                      right
        |                  |              |                  |                          |
        |               Literal           |               Literal                    Literal
        |          ________|______        |          ________|_______           ________|______
    data_type data_type         value data_type data_type          value   data_type         value
        |         |               |       |         |                |         |               |
     INTEGER   INTEGER            1    INTEGER   INTEGER             2      INTEGER            3

A simple multiply and add is represented as nodes in a tree, each node having a data type (here, all of them are
INTEGER).

A myriad of python data types are supported to convert to :class:`Literal` nodes using :func:`wrap_literal`. We even
support ``decimal.Decimal`` types for fixed precision decimal point classes in python. A notable exception is NoneType,
which cannot be wrapped to a literal. Instead, please use the :data:`NULL` singleton.
"""
import functools
from abc import ABC
from decimal import Decimal
from typing import Any, Dict, List, Optional, Type, TypeVar, Union

import attr

# NOTE: This is done so sphinx-autodoc-typehints doesn't run into a circular import issue with InQuery.
import treeno
from treeno.base import PrintMode, PrintOptions, Sql
from treeno.datatypes import types as type_consts
from treeno.datatypes.builder import (
    array,
    boolean,
    interval,
    row,
    time,
    timestamp,
    unknown,
)
from treeno.datatypes.conversions import get_arithmetic_type
from treeno.datatypes.inference import (
    infer_char,
    infer_decimal,
    infer_timelike_precision,
    infer_type,
)
from treeno.datatypes.types import DataType
from treeno.printer import join_stmts, pad
from treeno.util import (
    chain_identifiers,
    children,
    construct_container,
    is_dictlike,
    is_listlike,
    parenthesize,
    quote_literal,
)

GenericValue = TypeVar("GenericValue", bound="Value")

# Attr tries to assign __le__, __ge__, __eq__ and friends by default. We define our own.
value_attr = functools.partial(attr.s, order=False, eq=False, str=False)


@value_attr
class Value(Sql, ABC):
    """Represents a basic SQL value.

    A value can be one of the following:

    1. (:class:`Literal`) A literal value with well-defined type
    2. (:class:`Field`) A reference to a field in a table, which doesn't always have a well-defined type before
        resolution.
    3. (:class:`Star`) A reference to ALL of the fields in a table.
    4. (:class:`AliasedValue` and :class:`AliasedStar`) Aliased versions of (2) and (3).
    5. (:class:`Expression`) A nested complex expression involving any :class:`Value`. This can be a
        built-in operator such as +, a lambda expression, a complex function involving potentially variadic arguments
        as seen in :mod:`treeno.functions` inheriting from :class:`treeno.functions.base.Function`, etc.

    Attributes:
        data_type: The data type of the value. If not specified, defaults to UNKNOWN, which means the data type can't
            be directly determined.
    """

    data_type: DataType = attr.ib(factory=unknown, kw_only=True)

    def __invert__(self):
        return Not(self)

    def __pos__(self):
        return Positive(self)

    def __neg__(self):
        return Negative(self)

    def __add__(self, other):
        return Add(self, other)

    def __sub__(self, other):
        return Minus(self, other)

    def __mul__(self, other):
        return Multiply(self, other)

    def __truediv__(self, other):
        return Divide(self, other)

    def __pow__(self, other):
        from treeno.functions.math import Power

        return Power(self, other)

    def __mod__(self, other):
        return Modulus(self, other)

    def __eq__(self, other):
        return Equal(self, other)

    def __ne__(self, other):
        return NotEqual(self, other)

    def __gt__(self, other):
        return GreaterThan(self, other)

    def __ge__(self, other):
        return GreaterThanOrEqual(self, other)

    def __lt__(self, other):
        return LessThan(self, other)

    def __le__(self, other):
        return LessThanOrEqual(self, other)

    def __and__(self, other):
        return And(self, other)

    def __or__(self, other):
        return Or(self, other)

    def identifier(self) -> Optional[str]:
        """Identifier for the value.

        Values if aliased can have identifiers which are useful to capture output information. By default,
        :class:`Value` s are not identifiable by themselves.

        Returns:
            String name for the value's identifier if one exists, otherwise None.
        """
        return None


@value_attr
class Expression(Value, ABC):
    """Represents a complex expression which involves a function and its corresponding arguments.

    Expressions can be grouped into two general groups - :class:`treeno.functions.base.Function` s and built-in SQL operators. For all
    functions, please refer to the module :mod:`treeno.functions`. Operators live in `treeno.expression`, and support
    basic arihmetic scalar transforms such as add, subtract, divide, etc. and boolean clauses such as LIKE, IS NULL,
    DISTINCT FROM, etc.
    """


@value_attr
class Literal(Value):
    """Represents a literal value in SQL.

    Literal values are expressions that don't require calling constructors to any SQL data types such as ARRAY,
    and have a known data type (with the exception of NULL).

    >>> from decimal import Decimal
    >>> # Don't directly create a Literal value
    >>> Literal("a").data_type
    Traceback (most recent call last):
        ...
    AssertionError: Please use wrap_literal to construct a Literal value so the data types are detected
    >>> print(wrap_literal("a").data_type)
    VARCHAR(1)
    >>> print(wrap_literal(True).data_type)
    BOOLEAN
    >>> print(wrap_literal(1).data_type)
    INTEGER
    >>> print(wrap_literal(100000000000).data_type)
    BIGINT
    >>> print(wrap_literal(2.2).data_type)
    DOUBLE
    >>> # Use decimal.Decimal if you wish to represent a fixed precision decimal point in SQL
    >>> print(wrap_literal(Decimal("2.2")).data_type)
    DECIMAL(2,1)

    Attributes:
        value: A value in python representing a value in SQL.
    """

    value: Any = attr.ib()

    def __attrs_post_init__(self) -> None:
        if self.value is not None:
            assert (
                self.data_type != unknown()
            ), "Please use wrap_literal to construct a Literal value so the data types are detected"

    def sql(self, opts: PrintOptions) -> str:
        # NOTE: Literal decimals can be directly convertible
        # through string representation, i.e. 3.14 is DECIMAL(3,2), so we don't need to do anything.
        s = str(self.value)
        if self.value is None:
            s = "NULL"
        if isinstance(self.value, bool):
            s = s.upper()
        if isinstance(self.value, str):
            # Single quotes to mean literal string
            s = quote_literal(self.value)
        return s


@value_attr
class Field(Value):
    """Represents a field referenced in the input relations of a SELECT query

    Attributes:
        name: The name of the field
        table: The source of the field. This is sometimes necessary when there are multiple relations with duplicate
            column names and the name is not sufficient to disambiguate the field.
    """

    name: str = attr.ib()
    table: Optional[Union[str, Value]] = attr.ib(default=None)

    def __attrs_post_init__(self):
        assert not isinstance(
            self.table, Field
        ), "Table must be either a complex expression or a string representing the table, not a field"

    def identifier(self) -> Optional[str]:
        return self.name

    def sql(self, opts: PrintOptions) -> str:
        table_sql = None
        if self.table:
            if isinstance(self.table, str):
                table_sql = self.table
            else:
                table_sql = self.table.sql(opts)
        return chain_identifiers(table_sql, self.name)


@value_attr
class AliasedValue(Value):
    """Represents an alias on a value. For unpacking individual column aliases
    from a star, see AliasedStar
    """

    value: Value = attr.ib()
    alias: str = attr.ib()

    def __attrs_post_init__(self) -> None:
        assert not isinstance(
            self.value, Star
        ), "Stars cannot have aliases. Consider using AliasedStar"
        self.data_type = self.value.data_type

    def identifier(self) -> Optional[str]:
        return self.alias

    def sql(self, opts: PrintOptions) -> str:
        return f'{self.value.sql(opts)} "{self.alias}"'


@value_attr
class Star(Value):
    """Represents a `*` or a `table.*` statement

    Note:
        The reason Star does not inherit from Field is because a star has no name.
        Fields must have a name, and allow an optional table identifier.
    """

    # TODO: It should be noted that if table is a Field, that is a field referring to a table, not a column.
    table: Optional[Union[str, Value]] = attr.ib(default=None)

    def __attrs_post_init__(self):
        assert not isinstance(
            self.table, Field
        ), "Table must be either a complex expression or a string representing the table, not a field"

    def sql(self, opts: PrintOptions) -> str:
        star_string = ""
        if self.table:
            if isinstance(self.table, str):
                table_sql = self.table
            else:
                table_sql = self.table.sql(opts)
            star_string = f"{table_sql}."
        star_string += "*"
        return star_string


@value_attr
class AliasedStar(Value):
    """Represents one or more aliases corresponding to an unpacked star
    """

    star: Star = attr.ib()
    aliases: List[str] = attr.ib()

    def __attrs_post_init__(self) -> None:
        self.data_type = self.star.data_type

    def sql(self, opts: PrintOptions) -> str:
        alias_str = join_stmts(self.aliases, opts)
        return f"{self.star.sql(opts)} ({alias_str})"


@value_attr
class Lambda(Expression):
    """Represents an anonymous function. This expression will currently have an unknown type because it can never be
    expressed as a standalone expression.
    """

    @value_attr
    class Variable(Value):
        name: str = attr.ib()

        def sql(self, opts: PrintOptions) -> str:
            return self.name

        def identifier(self) -> Optional[str]:
            return self.name

    inputs: List[Variable] = attr.ib()
    expr: Value = attr.ib()

    def __attrs_post_init__(self):
        assert len(set(input.name for input in self.inputs)) == len(self.inputs)

    @classmethod
    def from_generic_expr(
        cls, inputs: List[Variable], expr: GenericValue
    ) -> "Lambda":
        input_set = set(input.name for input in inputs)

        def _from_expr(node: Any) -> Any:
            if is_listlike(node):
                return construct_container(
                    node, iter(_from_expr(child) for child in node)
                )
            if is_dictlike(node):
                return construct_container(
                    node,
                    iter((k, _from_expr(child)) for k, child in node.items()),
                )
            if not isinstance(node, Sql):
                return node
            if (
                isinstance(node, Field)
                and node.table is None
                and node.name in input_set
            ):
                return cls.Variable(node.name)
            changes = {}
            for k, v in children(node).items():
                changes[k] = _from_expr(v)
            return attr.evolve(node, **changes)

        return cls(inputs, _from_expr(expr))

    def sql(self, opts: PrintOptions) -> str:
        input_string = join_stmts([var.sql(opts) for var in self.inputs], opts)
        if len(self.inputs) > 1:
            input_string = parenthesize(input_string)
        return f"{input_string} -> {self.expr.sql(opts)}"


def wrap_literal(val: Any) -> Value:
    """Convenience method to wrap a literal value into a treeno Value"""
    if isinstance(val, Value):
        return val
    if val is None:
        raise ValueError(
            "Treeno explicitly disallows creating a None literal to avoid ambiguity. Please use treeno.expression.NULL instead."
        )
    assert not isinstance(
        val, (list, tuple, set, dict)
    ), "wrap_literal should not be used with composable types like ARRAY/MAP/ROW"
    return Literal(val, data_type=infer_type(val))


def wrap_literal_list(vals: List[Any]) -> List[Value]:
    return [wrap_literal(val) for val in vals]


def pemdas_str(
    current_type: Type[Value],
    val: Value,
    opts: PrintOptions,
    is_left: bool = True,
) -> str:
    """Apply parenthesization onto the nested expression if required for pemdas.

    Args:
        current_type: Current value's type.
        val: One of the direct children of the current value.
        is_left: Whether this value is the leftmost child. This is used to determine whether the right child should
            be parenthesized or not. i.e. 1+(2+3) or 1+2+3.
    """
    # If the type is not specified, we assume the syntax is well-formed without any parentheses as it's probably a fn
    # call.
    val_type = type(val)
    if (
        current_type not in OPERATOR_PRECEDENCE
        or val_type not in OPERATOR_PRECEDENCE
    ):
        return val.sql(opts)
    current_precedence = OPERATOR_PRECEDENCE[current_type]
    val_precedence = OPERATOR_PRECEDENCE[val_type]
    # The expression's precedence is the same, so we must obey left-to-right ordering.
    if current_precedence == val_precedence:
        if is_left:
            return val.sql(opts)
        else:
            return parenthesize(val)

    # The underlying value's precedence is lower, BUT it's deeper in the tree, which means we need to parenthesize it.
    if val_precedence < current_precedence:
        return parenthesize(val)
    return val.sql(opts)


@value_attr
class BinaryExpression(Expression, ABC):
    left: Value = attr.ib(converter=wrap_literal)
    right: Value = attr.ib(converter=wrap_literal)


@value_attr
class UnaryExpression(Expression, ABC):
    value: Value = attr.ib(converter=wrap_literal)


def builtin_binary_str(
    val: BinaryExpression, string_format: str, opts: PrintOptions
) -> str:
    return string_format.format(
        left=pemdas_str(type(val), val.left, opts, is_left=True),
        right=pemdas_str(type(val), val.right, opts, is_left=False),
    )


def builtin_unary_str(
    val: UnaryExpression, string_format: str, opts: PrintOptions
) -> str:
    return string_format.format(
        value=pemdas_str(type(val), val.value, opts, is_left=True)
    )


@value_attr
class Positive(UnaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = self.value.data_type

    def sql(self, opts: PrintOptions) -> str:
        return builtin_unary_str(self, "+{value}", opts)


@value_attr
class Negative(UnaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = self.value.data_type

    def sql(self, opts: PrintOptions) -> str:
        return builtin_unary_str(self, "-{value}", opts)


@value_attr
class Add(BinaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = get_arithmetic_type(
            self.left.data_type, self.right.data_type
        )

    def sql(self, opts: PrintOptions) -> str:
        return builtin_binary_str(self, "{left} + {right}", opts)


@value_attr
class Minus(BinaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = get_arithmetic_type(
            self.left.data_type, self.right.data_type
        )
        # Special minus operators for datetimes
        left_name, right_name = (
            self.left.data_type.type_name,
            self.right.data_type.type_name,
        )
        if {left_name, right_name}.issubset(
            {type_consts.TIMESTAMP, type_consts.DATE}
        ) or left_name == right_name == type_consts.TIME:
            self.data_type = interval(from_interval="DAY", to_interval="SECOND")

    def sql(self, opts: PrintOptions) -> str:
        return builtin_binary_str(self, "{left} - {right}", opts)


@value_attr
class Multiply(BinaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = get_arithmetic_type(
            self.left.data_type, self.right.data_type
        )

    def sql(self, opts: PrintOptions) -> str:
        return builtin_binary_str(self, "{left} * {right}", opts)


@value_attr
class Divide(BinaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = get_arithmetic_type(
            self.left.data_type, self.right.data_type
        )

    def sql(self, opts: PrintOptions) -> str:
        return builtin_binary_str(self, "{left} / {right}", opts)


@value_attr
class Not(UnaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def sql(self, opts: PrintOptions) -> str:
        # Specializations on Not
        if isinstance(
            self.value,
            (
                DistinctFrom,
                IsNull,
                Like,
                InList,
                InQuery,
                Between,
                Equal,
                NotEqual,
            ),
        ):
            return self.value.to_string(opts, negate=True)

        return builtin_unary_str(self, "NOT {value}", opts)


@value_attr
class Modulus(BinaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = get_arithmetic_type(
            self.left.data_type, self.right.data_type
        )

    def sql(self, opts: PrintOptions) -> str:
        return builtin_binary_str(self, "{left} % {right}", opts)


@value_attr
class Equal(BinaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def to_string(self, opts: PrintOptions, negate: bool = False) -> str:
        if negate:
            return NotEqual(self.left, self.right).sql(opts)
        return builtin_binary_str(self, "{left} = {right}", opts)

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string(opts, negate=False)


@value_attr
class NotEqual(BinaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def to_string(self, opts: PrintOptions, negate: bool = False) -> str:
        if negate:
            return str(Equal(self.left, self.right))
        return builtin_binary_str(self, "{left} <> {right}", opts)

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string(opts, negate=False)


@value_attr
class GreaterThan(BinaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def sql(self, opts: PrintOptions) -> str:
        return builtin_binary_str(self, "{left} > {right}", opts)


@value_attr
class GreaterThanOrEqual(BinaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def sql(self, opts: PrintOptions) -> str:
        return builtin_binary_str(self, "{left} >= {right}", opts)


@value_attr
class LessThan(BinaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def sql(self, opts: PrintOptions) -> str:
        return builtin_binary_str(self, "{left} < {right}", opts)


@value_attr
class LessThanOrEqual(BinaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def sql(self, opts: PrintOptions) -> str:
        return builtin_binary_str(self, "{left} <= {right}", opts)


@value_attr
class And(BinaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def sql(self, opts: PrintOptions) -> str:
        # According to the specifications, AND and OR should be on the next line if we're in pretty mode
        spacing = "\n" if opts.mode == PrintMode.PRETTY else " "
        return builtin_binary_str(
            self, "{left}" + spacing + "AND {right}", opts
        )


@value_attr
class Or(BinaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def sql(self, opts: PrintOptions) -> str:
        spacing = "\n" if opts.mode == PrintMode.PRETTY else " "
        return builtin_binary_str(self, "{left}" + spacing + "OR {right}", opts)


@value_attr
class IsNull(UnaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def to_string(self, opts: PrintOptions, negate: bool = False):
        return builtin_unary_str(
            self, "{value} IS NOT NULL" if negate else "{value} IS NULL", opts
        )

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string(opts, negate=False)


@value_attr
class DistinctFrom(BinaryExpression):
    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def to_string(self, opts: PrintOptions, negate: bool = False):
        return builtin_binary_str(
            self,
            "{left} IS NOT DISTINCT FROM {right}"
            if negate
            else "{left} IS DISTINCT FROM {right}",
            opts,
        )

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string(opts, negate=False)


@value_attr
class Between(Expression):
    value: Value = attr.ib(converter=wrap_literal)
    lower: Value = attr.ib(converter=wrap_literal)
    upper: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def to_string(self, opts: PrintOptions, negate: bool = False) -> str:
        between_string = pemdas_str(type(self), self.value, opts)
        if negate:
            between_string += " NOT"
        between_string += f" BETWEEN {pemdas_str(type(self), self.lower, opts)} AND {pemdas_str(type(self), self.upper, opts)}"
        return between_string

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string(opts, negate=False)


@value_attr
class Array(Expression):
    values: List[Value] = attr.ib(converter=wrap_literal_list)

    def __attrs_post_init__(self) -> None:
        assert len(self.values), "values must be a non-empty list for Array"
        # TODO: This is currently incorrect. We have to widen the data type depending on the largest precision we have
        # i.e. TYPEOF(ARRAY[CHAR '3'] || '345') yields array(char(3))
        self.data_type = array(dtype=self.values[0].data_type)

    @classmethod
    def from_values(cls, *vals: Any) -> "Array":
        return cls(vals)

    def sql(self, opts: PrintOptions) -> str:
        values_str = join_stmts([val.sql(opts) for val in self.values], opts)
        return f"ARRAY[{values_str}]"


@value_attr
class InQuery(Expression):
    value: Value = attr.ib(converter=wrap_literal)
    query: "treeno.relation.Query" = attr.ib()

    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def to_string(self, opts: PrintOptions, negate: bool = False) -> str:
        in_list_string = pemdas_str(type(self), self.value, opts)
        if negate:
            in_list_string += " NOT"
        in_list_string += f" IN ({self.query.sql(opts)})"
        return in_list_string

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string(opts, negate=False)


@value_attr
class InList(Expression):
    value: Value = attr.ib(converter=wrap_literal)
    exprs: List[Value] = attr.ib(converter=wrap_literal_list)

    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def to_string(self, opts: PrintOptions, negate: bool = False) -> str:
        expr_list = join_stmts([expr.sql(opts) for expr in self.exprs], opts)
        in_list_string = pemdas_str(type(self), self.value, opts)
        if negate:
            in_list_string += " NOT"
        in_list_string += f" IN ({expr_list})"
        return in_list_string

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string(opts, negate=False)


@value_attr
class Like(Expression):
    value: Value = attr.ib(converter=wrap_literal)
    pattern: Value = attr.ib(converter=wrap_literal)
    escape: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )

    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()

    def to_string(self, opts: PrintOptions, negate: bool = False) -> str:
        like_str = pemdas_str(type(self), self.value, opts)
        if negate:
            like_str += " NOT"
        like_str += f" LIKE {pemdas_str(type(self), self.pattern, opts)}"
        if self.escape:
            # TODO: Should we pemdas this? I don't see it in the specs
            like_str += f" ESCAPE {self.escape.sql(opts)}"
        return like_str

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string(opts, negate=False)


@value_attr
class TypeConstructor(Expression):
    value: str = attr.ib()
    type_name: str = attr.ib(converter=str.upper)

    def __attrs_post_init__(self) -> None:
        # This can't be timezoned, so we ignore that
        if self.type_name == type_consts.TIMESTAMP:
            self.data_type = timestamp(
                precision=infer_timelike_precision(self.value)
            )
        if self.type_name == type_consts.TIME:
            self.data_type = time(
                precision=infer_timelike_precision(self.value)
            )
        elif self.type_name == type_consts.DECIMAL:
            self.data_type = infer_decimal(Decimal(self.value))
        elif self.type_name == type_consts.CHAR:
            self.data_type = infer_char(self.value)
        else:
            self.data_type = DataType(self.type_name)

    def sql(self, opts: PrintOptions) -> str:
        # We aren't allowed to parametrize the types here.
        # There's an edge case where parametrizing timestamp with timezone=True is not in the parens notation
        # but that also fails for this(since the parser recognizes a single string as the type) which means we
        # can just take the type name.
        return f"{self.type_name} {quote_literal(self.value)}"


@value_attr
class RowConstructor(Expression):
    values: List[Value] = attr.ib()

    def __attrs_post_init__(self) -> None:
        self.data_type = row(dtypes=[value.data_type for value in self.values])

    def sql(self, opts: PrintOptions) -> str:
        values_string = join_stmts(
            [value.sql(opts) for value in self.values], opts
        )
        return f"({values_string})"


@value_attr
class Interval(Expression):
    value: str = attr.ib()
    from_interval: str = attr.ib()
    to_interval: Optional[str] = attr.ib(default=None)

    def __attrs_post_init__(self) -> None:
        if self.from_interval in ("YEAR", "MONTH"):
            self.data_type = interval(from_interval="YEAR", to_interval="MONTH")
        else:
            assert self.from_interval in (
                "DAY",
                "HOUR",
                "MINUTE",
                "SECOND",
            ), f"Unknown interval type {self.from_interval}"
            self.data_type = interval(from_interval="DAY", to_interval="SECOND")

    def sql(self, opts: PrintOptions) -> str:
        to_interval_str = f" TO {self.to_interval}" if self.to_interval else ""
        return f"INTERVAL {self.value} {self.from_interval}" + to_interval_str


@value_attr
class Cast(Expression):
    expr: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        # TODO: We should probably just set init=False to attr.s and force them to input the data type as a positional
        # argument. This is a temporary workaround.
        assert self.data_type != unknown(), "data_type must be defined for Cast"

    def sql(self, opts: PrintOptions) -> str:
        return f"CAST({self.expr.sql(opts)} AS {self.data_type.sql(opts)})"


@value_attr
class TryCast(Expression):
    expr: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self):
        assert (
            self.data_type != unknown()
        ), "data_type must be defined for TryCast"

    def sql(self, opts: PrintOptions) -> str:
        return f"TRY_CAST({self.expr.sql(opts)} AS {self.data_type.sql(opts)})"


@value_attr
class Subscript(Expression):
    value: Value = attr.ib()
    index: Value = attr.ib(converter=wrap_literal)

    def sql(self, opts: PrintOptions) -> str:
        return f"{self.value.sql(opts)}[{self.index.sql(opts)}]"


# These are not values by themselves and must be used in the context of `Case`.
@attr.s
class When(Sql):
    condition: Value = attr.ib(converter=wrap_literal)
    value: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        self.data_type = self.value.data_type

    def sql(self, opts: PrintOptions):
        return f"WHEN {self.condition.sql(opts)} THEN {self.value.sql(opts)}"


@attr.s
class Else(Sql):
    value: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        self.data_type = self.value.data_type

    def sql(self, opts: PrintOptions):
        return f"ELSE {self.value.sql(opts)}"


@value_attr
class Case(Expression):
    branches: List[When] = attr.ib()
    else_: Optional[Else] = attr.ib(default=None)
    value: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )

    def __attrs_post_init__(self) -> None:
        assert len(
            self.branches
        ), "There must be at least one WHEN condition in a CASE"
        self.data_type = self.branches[0].data_type

    def sql(self, opts: PrintOptions) -> str:
        spacing = "\n" if opts.mode == PrintMode.PRETTY else " "
        conds = self.branches
        if self.else_:
            conds = conds + [self.else_]
        branches_string = pad(
            spacing + spacing.join([cond.sql(opts) for cond in conds]), 4
        )
        end_string = spacing + "END"
        value_string = self.value.sql(opts) if self.value is not None else ""
        return f"CASE {value_string}{branches_string}{end_string}"


# Operator precedence according to:
# https://docs.oracle.com/cd/B19306_01/server.102/b14200/operators001.htm
# https://docs.oracle.com/cd/B19306_01/server.102/b14200/conditions001.htm#i1034834
# Still waiting to hear back about Trino's custom precedence:
# https://trinodb.slack.com/archives/CFLB9AMBN/p1637528834018300
OPERATOR_PRECEDENCE: Dict[Type[Value], int] = {
    Positive: 8,
    Negative: 8,
    Multiply: 7,
    Divide: 7,
    Modulus: 7,
    Add: 6,
    Minus: 6,
    # TODO: Currently we run into an import loop trying to add Concatenate here, so we add it at the end in functions.common
    Equal: 4,
    NotEqual: 4,
    GreaterThan: 4,
    GreaterThanOrEqual: 4,
    LessThan: 4,
    LessThanOrEqual: 4,
    IsNull: 3,
    Like: 3,
    Between: 3,
    InList: 3,
    InQuery: 3,
    Not: 2,
    And: 1,
    Or: 0,
}

# Use this instead of wrap_literal(None), which we explicitly disallow
NULL = Literal(None, data_type=unknown())



"""A printer that follows a partial set of SQL formatting standards.
StatementPrinter complies with sqlstyle.guide.

Note that none of these printers are explicitly responsible for calling sql(). They should work only with the
primitive str representation.
"""
from typing import Dict, List

import attr

from treeno.base import PrintMode, PrintOptions


def pad(input: str, spaces: int) -> str:
    """Pads the input lines with spaces except for the first line.
    We don't need opts here because the default mode will always be a single line so we'll never invoke this.
    """
    lines = input.splitlines()
    pad = " " * spaces
    return f"\n{pad}".join(lines)


def join_stmts(
    stmts: List[str], opts: PrintOptions, delimiter: str = ","
) -> str:
    """A convenience method for joining statements when the entire stmt list is already known"""
    return JoinPrinter(delimiter=delimiter, stmt_list=stmts).to_string(opts)


@attr.s
class JoinPrinter:
    """JoinPrinter is responsible for formatting a sequence of strings. For example:

    SELECT abc, 123, -- gets wrapped over
           def, ghij
    """

    delimiter: str = attr.ib()
    stmt_list: List[str] = attr.ib(factory=list)
    max_length: int = attr.ib(default=80)

    def add_entry(self, value: str) -> None:
        self.stmt_list.append(value)

    def update(self, values: List[str]) -> None:
        self.stmt_list.extend(values)

    def to_string(self, opts: PrintOptions) -> str:
        assert opts.spaces < self.max_length
        remaining_length = self.max_length - opts.spaces
        lines = []
        current_length = 0
        for idx, line in enumerate(self.stmt_list):
            # If we're at the last element, we don't need to add a comma
            needs_comma = int(idx < len(self.stmt_list) - 1)
            line_length = len(line) + needs_comma
            # If a line is simply too long, we can't do anything about it but to include it in its own line.
            if (
                opts.mode == PrintMode.PRETTY
                and current_length + line_length > remaining_length
            ):
                current_length = line_length
                # If we're at the beginning of the line, we shouldn't add a newline
                newline_if_not_beginning = "\n" if idx != 0 else ""
                lines.append(newline_if_not_beginning + line)
            else:
                # Add 1 because of commas
                current_length += line_length
                lines.append(line)
        return self.delimiter.join(lines)


@attr.s
class StatementPrinter:
    """StatementPrinter is responsible for formatting large hierarchical statements that usually consist of consecutive
    lines of <keyword> <sql expression>. For example:

    SELECT model_num
      FROM phones AS p
     WHERE p.release_date > '2014-09-30';

    The keywords are right adjusted to form a "river" for better readability
    """

    # Buffer is a logical per-statement buffer which tries to right-adjust the keywords and left-adjust the sql entities
    stmt_mapping: Dict[str, str] = attr.ib(factory=dict)
    river: bool = attr.ib(default=True)

    def add_entry(self, key: str, value: str) -> None:
        """Adds an entry to the statement.
        Note that if value is an empty string, we will omit it from the formatting later on.
        """
        assert (
            key not in self.stmt_mapping
        ), f"Key {key} already exists in statement printer"
        self.stmt_mapping[key] = value

    def update(self, new_mapping: Dict[str, str]) -> None:
        existing_keys = new_mapping.keys() & self.stmt_mapping.keys()
        assert not existing_keys, f"Keys {existing_keys} already exist"
        for key, value in new_mapping.items():
            self.add_entry(key, value)

    def to_string(self, opts: PrintOptions) -> str:
        if not self.stmt_mapping:
            return ""
        rpad = max(len(line) for line in self.stmt_mapping)
        lines = []
        for keyword, sql_str in self.stmt_mapping.items():
            if opts.mode == PrintMode.PRETTY:
                # If we don't care about the river, then don't pad anything.
                if self.river:
                    keyword = keyword.rjust(rpad)
                    # If sql_str is empty, then no need to pad.
                    if sql_str:
                        sql_str = pad(sql_str, rpad + 1)
            elif opts.mode != PrintMode.DEFAULT:
                raise NotImplementedError(
                    f"to_string not implemented for mode {opts.mode}"
                )

            if sql_str:
                lines.append(keyword + " " + sql_str)
            else:
                lines.append(keyword)

        join_char = "\n" if opts.mode == PrintMode.PRETTY else " "
        return join_char.join(lines)



import os
import subprocess
from setuptools import setup
from setuptools.command.develop import develop


class AntlrDevelopCommand(develop):
    def run(self):
        output_dir = compile_grammar()
        print(f"Compiled ANTLRv4 grammar in {output_dir}")
        develop.run(self)


def compile_grammar():
    parser_dir = os.path.join(os.path.dirname(__file__), "treeno/grammar")
    subprocess.check_output(
        ["antlr", "SqlBase.g4", "-Dlanguage=Python3", "-visitor", "-o", "gen"],
        cwd=parser_dir,
    )
    # The result is created in the subfolder `gen`
    return os.path.join(parser_dir, "gen")


setup(cmdclass={"develop": AntlrDevelopCommand})
